// *******************************************************************
//   (C) Copyright 2014 Leiden Institute of Advanced Computer Science
//   Universiteit Leiden
//   All Rights Reserved
// *******************************************************************
// Tablut (core)
// *******************************************************************
// FILE INFORMATION:
//   File:     position.cc (depends on position.h)
//   Author:   Jonathan K. Vis
//   Revision: 1.01a
//   Date:     2014/01/20
// *******************************************************************
// DESCRIPTION:
//   Defines the position object which contains a (legal) game
//   configuration.
// *******************************************************************

#include "position.h"

int Position::generate_moves(Move moves[MAX_MOVES]) const
{
  int count = 0;
  if (is_white_turn())
  {
    for (int piece = WHITE_KING; piece <= WHITE_PAWN_7; ++piece)
    {
      count += generate_moves(moves + count, piece);
    } // for
  } // if
  else
  {
    for (int piece = BLACK_PAWN_0; piece <= BLACK_PAWN_15; ++piece)
    {
      count += generate_moves(moves + count, piece);
    } // for
  } // else
  return count;
} // Position::generate_moves

int Position::generate_moves(Move moves[MAX_MOVES_PER_PIECE], int const piece) const
{
  int count = 0;
  if (pieces[piece] == CAPTURED)
  {
    return count;
  } // if
  // moving north
  for (int rank = pieces[piece].rank - 1; rank >= 0 && board[rank][pieces[piece].file] == NONE; --rank)
  {
    if (Square(pieces[piece].file, rank) != CASTLE)
    {
      moves[count] = Move(pieces[piece], Square(pieces[piece].file, rank));
      generate_captures(moves[count]);
      ++count;
    } // if
  } // for
  // moving east
  for (int file = pieces[piece].file + 1; file < 9 && board[pieces[piece].rank][file] == NONE; ++file)
  {
    if (Square(file, pieces[piece].rank) != CASTLE)
    {
      moves[count] = Move(pieces[piece], Square(file, pieces[piece].rank));
      generate_captures(moves[count]);
      ++count;
    } // if
  } // for
  // moving south
  for (int rank = pieces[piece].rank + 1; rank < 9 && board[rank][pieces[piece].file] == NONE; ++rank)
  {
    if (Square(pieces[piece].file, rank) != CASTLE)
    {
      moves[count] = Move(pieces[piece], Square(pieces[piece].file, rank));
      generate_captures(moves[count]);
      ++count;
    } // if
  } // for
  // moving west
  for (int file = pieces[piece].file - 1; file >= 0 && board[pieces[piece].rank][file] == NONE; --file)
  {
    if (Square(file, pieces[piece].rank) != CASTLE)
    {
      moves[count] = Move(pieces[piece], Square(file, pieces[piece].rank));
      generate_captures(moves[count]);
      ++count;
    } // if
  } // for
  return count;
} // Position::generate_moves


void Position::generate_captures(Move &move) const
{
  if (move.from.rank >= move.to.rank && move.to.rank > 1 &&
      is_friendly_pawn(board[move.from.rank][move.from.file], board[move.to.rank - 2][move.to.file]) &&
      is_enemy_pawn(board[move.from.rank][move.from.file], board[move.to.rank - 1][move.to.file]))
  {
    move.capture[0] = board[move.to.rank - 1][move.to.file];
  } // if
  else
  {
    move.capture[0] = NONE;
  } // else
  if (move.from.file <= move.to.file && move.to.file < 7 &&
      is_friendly_pawn(board[move.from.rank][move.from.file], board[move.to.rank][move.to.file + 2]) &&
      is_enemy_pawn(board[move.from.rank][move.from.file], board[move.to.rank][move.to.file + 1]))
  {
    move.capture[1] = board[move.to.rank][move.to.file + 1];
  } // if
  else
  {
    move.capture[1] = NONE;
  } // else
  if (move.from.rank <= move.to.rank && move.to.rank < 7 &&
      is_friendly_pawn(board[move.from.rank][move.from.file], board[move.to.rank + 2][move.to.file]) &&
      is_enemy_pawn(board[move.from.rank][move.from.file], board[move.to.rank + 1][move.to.file]))
  {
    move.capture[2] = board[move.to.rank + 1][move.to.file];
  } // if
  else
  {
    move.capture[2] = NONE;
  } // else
  if (move.from.file >= move.to.file && move.to.file > 1 &&
      is_friendly_pawn(board[move.from.rank][move.from.file], board[move.to.rank][move.to.file - 2]) &&
      is_enemy_pawn(board[move.from.rank][move.from.file], board[move.to.rank][move.to.file - 1]))
  {
    move.capture[3] = board[move.to.rank][move.to.file - 1];
  } // if
  else
  {
    move.capture[3] = NONE;
  } // else
  return;
} // Position::generate_captures

bool Position::is_castle_capture(void) const
{
  if (is_black_pawn(board[last_move.rank][last_move.file]))
  {
    if (last_move == E3)
    {
      return pieces[WHITE_KING] == E4 && is_black_pawn(board[F4.rank][F4.file]) && is_black_pawn(board[D4.rank][D4.file]);
    } // if
    if (last_move == F4)
    {
      return (pieces[WHITE_KING] == E4 && is_black_pawn(board[E3.rank][E3.file]) && is_black_pawn(board[D4.rank][D4.file])) ||
             (pieces[WHITE_KING] == F5 && is_black_pawn(board[G5.rank][G5.file]) && is_black_pawn(board[F6.rank][F6.file]));
    } // if
    if (last_move == G5)
    {
      return pieces[WHITE_KING] == F5 && is_black_pawn(board[F4.rank][F4.file]) && is_black_pawn(board[F6.rank][F6.file]);
    } // if
    if (last_move == F6)
    {
      return (pieces[WHITE_KING] == F5 && is_black_pawn(board[F4.rank][F4.file]) && is_black_pawn(board[G5.rank][G5.file])) ||
             (pieces[WHITE_KING] == E6 && is_black_pawn(board[E7.rank][E7.file]) && is_black_pawn(board[D6.rank][D6.file]));
    } // if
    if (last_move == E7)
    {
      return pieces[WHITE_KING] == E6 && is_black_pawn(board[F6.rank][F6.file]) && is_black_pawn(board[D6.rank][D6.file]);
    } // if
    if (last_move == D6)
    {
      return (pieces[WHITE_KING] == E6 && is_black_pawn(board[F6.rank][F6.file]) && is_black_pawn(board[E7.rank][E7.file])) ||
             (pieces[WHITE_KING] == D5 && is_black_pawn(board[C5.rank][C5.file]) && is_black_pawn(board[D4.rank][D4.file]));
    } // if
    if (last_move == C5)
    {
      return pieces[WHITE_KING] == D5 && is_black_pawn(board[D6.rank][D6.file]) && is_black_pawn(board[D4.rank][D4.file]);
    } // if
    if (last_move == D4)
    {
      return (pieces[WHITE_KING] == D5 && is_black_pawn(board[C5.rank][C5.file]) && is_black_pawn(board[D6.rank][D6.file])) ||
             (pieces[WHITE_KING] == E4 && is_black_pawn(board[E3.rank][E3.file]) && is_black_pawn(board[F4.rank][F4.file]));
    } // if
  } // if
  return false;
} // Position::king_captured

