// *******************************************************************
//   (C) Copyright 2014 Leiden Institute of Advanced Computer Science
//   Universiteit Leiden
//   All Rights Reserved
// *******************************************************************
// Tablut (core)
// *******************************************************************
// FILE INFORMATION:
//   File:     position.h (implemented in position.cc)
//   Author:   Jonathan K. Vis
//   Revision: 1.01a
//   Date:     2014/02/10
// *******************************************************************
// DESCRIPTION:
//   Defines the position object which contains a (legal) game
//   configuration.
// *******************************************************************

#if !defined(__position_h__)
#define __position_h__

#include "location.h"
#include "move.h"
#include "pieces.h"
#include "square.h"
#include "zobrist.h"


class Position
{
  public:
    explicit Position(void) { }


    void clear(void);
    void initial(void);


    int get_board(int const rank, int const file) const;
    Square get_piece(int const piece) const;
    int get_turn(void) const;
    uint64_t get_key(void) const;
    Square get_last_move(void) const;

    bool is_legal_move(Move &move) const;
    bool is_white_turn(void) const;
    bool is_won(void) const;


    void set_piece(int const piece, Square const &location);
    void set_key(uint64_t const key);
    void set_last_move(Square const last_move);


    int generate_moves(Move moves[MAX_MOVES]) const;
    int generate_moves(Move      moves[MAX_MOVES_PER_PIECE],
                       int const piece) const;
    void do_move(Move const &move);
    void undo_move(Move const &move);
    void do_null_move(void);
    void undo_null_move(void);

  private:
    void generate_captures(Move &move) const;
    bool is_castle_capture(void) const;


    int board[9][9];
    Square pieces[26];
    int turn;
    uint64_t key;

    Square last_move;

}; // Position


inline void Position::clear(void)
{
  for (int rank = 0; rank < 9; ++rank)
  {
    for (int file = 0; file < 9; ++file)
    {
      board[rank][file] = NONE;
    } // for
  } // for
  for (int piece = 0; piece < 26; ++piece)
  {
    pieces[piece] = CAPTURED;
  } // for
  turn = 0;
  key = 0x0ull;
  last_move = CASTLE;
  return;
} // Position::clear

inline void Position::initial(void)
{
  clear();
  set_piece(WHITE_KING, CASTLE);
  set_piece(WHITE_PAWN_0, E6);
  set_piece(WHITE_PAWN_1, F5);
  set_piece(WHITE_PAWN_2, E4);
  set_piece(WHITE_PAWN_3, D5);
  set_piece(WHITE_PAWN_4, E7);
  set_piece(WHITE_PAWN_5, G5);
  set_piece(WHITE_PAWN_6, E3);
  set_piece(WHITE_PAWN_7, C5);
  set_piece(BLACK_PAWN_0, E8);
  set_piece(BLACK_PAWN_1, H5);
  set_piece(BLACK_PAWN_2, E2);
  set_piece(BLACK_PAWN_3, B5);
  set_piece(BLACK_PAWN_4, D9);
  set_piece(BLACK_PAWN_5, E9);
  set_piece(BLACK_PAWN_6, F9);
  set_piece(BLACK_PAWN_7, I6);
  set_piece(BLACK_PAWN_8, I5);
  set_piece(BLACK_PAWN_9, I4);
  set_piece(BLACK_PAWN_10, D1);
  set_piece(BLACK_PAWN_11, E1);
  set_piece(BLACK_PAWN_12, F1);
  set_piece(BLACK_PAWN_13, A6);
  set_piece(BLACK_PAWN_14, A5);
  set_piece(BLACK_PAWN_15, A4);
  set_last_move(CASTLE);
  return;
} // Position::initial

inline int Position::get_board(int const rank, int const file) const
{
  return board[rank][file];
} // Position::get_board

inline Square Position::get_piece(int const piece) const
{
  return pieces[piece];
} // Position::get_piece

inline int Position::get_turn(void) const
{
  return turn;
} // Position::get_turn

inline uint64_t Position::get_key(void) const
{
  return key;
} // Position::get_key

inline Square Position::get_last_move(void) const
{
  return last_move;
} // Position::get_last_move

inline void Position::set_piece(int const     piece,
                                Square const &location)
{
  board[location.rank][location.file] = piece;
  pieces[piece] = location;
  key ^= ZOBRIST[piece][location.rank][location.file];
  return;
} // Position::set_piece

inline void Position::set_key(uint64_t const key)
{
  this->key = key;
  return;
} // Position::set_key

inline void Position::set_last_move(Square const last_move)
{
  this->last_move = last_move;
  key ^= ZOBRIST[0][last_move.rank][last_move.file];
  return;
} // Position::set_last_move

inline bool Position::is_legal_move(Move &move) const
{
  Move moves[MAX_MOVES];
  int const count = generate_moves(moves);
  for (int i = 0; i < count; ++i)
  {
    if (moves[i] == move)
    {
      move = moves[i];
      return true;
    } // if
  } // for
  return false;
} // Position::is_legal_move

inline bool Position::is_white_turn(void) const
{
  return turn % 2 == 0;
} // Position::is_white_turn

inline bool Position::is_won(void) const
{
  return pieces[WHITE_KING].rank == 0 ||
         pieces[WHITE_KING].file == 8 ||
         pieces[WHITE_KING].rank == 8 ||
         pieces[WHITE_KING].file == 0 ||
         (is_black_pawn(board[pieces[WHITE_KING].rank - 1][pieces[WHITE_KING].file]) &&
          is_black_pawn(board[pieces[WHITE_KING].rank][pieces[WHITE_KING].file + 1]) &&
          is_black_pawn(board[pieces[WHITE_KING].rank + 1][pieces[WHITE_KING].file]) &&
          is_black_pawn(board[pieces[WHITE_KING].rank][pieces[WHITE_KING].file - 1])) ||
         is_castle_capture();
} // Position::is_won

inline void Position::do_move(Move const &move)
{
  int const piece = board[move.from.rank][move.from.file];
  board[move.from.rank][move.from.file] = NONE;
  board[move.to.rank][move.to.file] = piece;
  pieces[piece] = move.to;
  key ^= ZOBRIST[piece][move.from.rank][move.from.file];
  key ^= ZOBRIST[piece][move.to.rank][move.to.file];
  if (move.capture[0] != NONE)
  {
    board[move.to.rank - 1][move.to.file] = NONE;
    pieces[move.capture[0]] = CAPTURED;
    key ^= ZOBRIST[move.capture[0]][move.to.rank - 1][move.to.file];
  } // if
  if (move.capture[1] != NONE)
  {
    board[move.to.rank][move.to.file + 1] = NONE;
    pieces[move.capture[1]] = CAPTURED;
    key ^= ZOBRIST[move.capture[1]][move.to.rank][move.to.file + 1];
  } // if
  if (move.capture[2] != NONE)
  {
    board[move.to.rank + 1][move.to.file] = NONE;
    pieces[move.capture[2]] = CAPTURED;
    key ^= ZOBRIST[move.capture[2]][move.to.rank + 1][move.to.file];
  } // if
  if (move.capture[3] != NONE)
  {
    board[move.to.rank][move.to.file - 1] = NONE;
    pieces[move.capture[3]] = CAPTURED;
    key ^= ZOBRIST[move.capture[3]][move.to.rank][move.to.file - 1];
  } // if
  ++turn;
  key ^= ZOBRIST_TURN;
  last_move = Square(move.to.file, move.to.rank);
  return;
} // Position::do_move

inline void Position::undo_move(Move const &move)
{
  int const piece = board[move.to.rank][move.to.file];
  board[move.to.rank][move.to.file] = NONE;
  board[move.from.rank][move.from.file] = piece;
  pieces[piece] = move.from;
  key ^= ZOBRIST[piece][move.to.rank][move.to.file];
  key ^= ZOBRIST[piece][move.from.rank][move.from.file];
  if (move.capture[0] != NONE)
  {
    board[move.to.rank - 1][move.to.file] = move.capture[0];
    pieces[move.capture[0]] = Square(move.to.file, move.to.rank - 1);
    key ^= ZOBRIST[move.capture[0]][move.to.rank - 1][move.to.file];
  } // if
  if (move.capture[1] != NONE)
  {
    board[move.to.rank][move.to.file + 1] = move.capture[1];
    pieces[move.capture[1]] = Square(move.to.file + 1, move.to.rank);
    key ^= ZOBRIST[move.capture[1]][move.to.rank][move.to.file + 1];
  } // if
  if (move.capture[2] != NONE)
  {
    board[move.to.rank + 1][move.to.file] = move.capture[2];
    pieces[move.capture[2]] = Square(move.to.file, move.to.rank + 1);
    key ^= ZOBRIST[move.capture[2]][move.to.rank + 1][move.to.file];
  } // if
  if (move.capture[3] != NONE)
  {
    board[move.to.rank][move.to.file - 1] = move.capture[3];
    pieces[move.capture[3]] = Square(move.to.file - 1, move.to.rank);
    key ^= ZOBRIST[move.capture[3]][move.to.rank][move.to.file - 1];
  } // if
  --turn;
  key ^= ZOBRIST_TURN;
  last_move = CASTLE;
  return;
} // Position::undo_move

inline void Position::do_null_move(void)
{
  ++turn;
  key ^= ZOBRIST_TURN;
  last_move = CASTLE;
  return;
} // Position::do_null_move

inline void Position::undo_null_move(void)
{
  --turn;
  key ^= ZOBRIST_TURN;
  last_move = CASTLE;
  return;
} // Position::undo_null_move

#endif

